


LOG_FILTER_TYPES = {}

def register_filter_type(name, klass):
    LOG_FILTER_TYPES[name] = klass

def unregister_filter_type(name):
    del LOG_FILTER_TYPES[name]

def get_filter_class(name):
    return LOG_FILTER_TYPES.get(name, None)
    
class LogFilter(object):

    def __init__(self, settings):
        self.settings = settings or {}

    def line_filter(self, line):
        raise NotImplementedError()

    @classmethod
    def init(cls, settings):
        filter_type = settings.get("type", "icontains")
        filter_class = get_filter_class(filter_type)
        if not filter_class:
            return None
        return filter_class(settings)


class ContainsFilter(LogFilter):

    def line_filter(self, line):
        substrings = self.settings.get("substrings", [])
        for substring in substrings:
            if not substring in line:
                return False
        return True

class IContainsFilter(LogFilter):

    def line_filter(self, line):
        line = line.lower()
        substrings = self.settings.get("substrings", [])
        for substring in substrings:
            if not substring.lower() in line:
                return False
        return True

register_filter_type("contains", ContainsFilter)
register_filter_type("icontains", IContainsFilter)



# -*- coding: utf-8 -*-

# Simple Type Data
INTEGER_VALUE = "([+|-]?\\d+)"
FLOAT_VALUE = "([+-]?\\d+\\.?\\d+)"
HEX_VALUE = "([0-9abcdefABCDEF]+)"
NOSPACE_VALUE = "([^\\s]+)"

# Special Type Data
VERSION_VALUE = "(\\d+(\\.\\d+){0,3})"
FQDN_VALUE = "([a-zA-Z\\d-]+(\\.[a-zA-Z\\d-]+)*)"
ABSOLUTE_PATH_VALUE = "((\\/[^/]+)+\\/?)"
OID_VALUE = "(\\d+(\\.\\d+)*)"
IPV4SEG  = "(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])"
IPV4ADDR = "(({IPV4SEG}\.){{3,3}}{IPV4SEG})".format(IPV4SEG=IPV4SEG)
IPV6SEG = "([0-9a-fA-F]{1,4})"
IPV6ADDR = """(
    ({IPV6SEG}:){{7,7}}{IPV6SEG}|
    ({IPV6SEG}:){{1,7}}:|
    ({IPV6SEG}:){{1,6}}:{IPV6SEG}|
    ({IPV6SEG}:){{1,5}}(:{IPV6SEG}){{1,2}}|
    ({IPV6SEG}:){{1,4}}(:{IPV6SEG}){{1,3}}|
    ({IPV6SEG}:){{1,3}}(:{IPV6SEG}){{1,4}}|
    ({IPV6SEG}:){{1,2}}(:{IPV6SEG}){{1,5}}|
    {IPV6SEG}:((:{IPV6SEG}){{1,6}})|
    :((:{IPV6SEG}){{1,7}}|:)|
    fe80:(:{IPV6SEG}){{0,4}}%[0-9a-zA-Z]{{1,}}|
    ::(ffff(:0{{1,4}}){{0,1}}:){{0,1}}{IPV4ADDR}|
    ({IPV6SEG}:){{1,4}}:{IPV4ADDR}
    )""".format(IPV6SEG=IPV6SEG, IPV4ADDR=IPV4ADDR).replace(" ", "").replace("\r", "").replace("\n", "")
IPV4 = IPV4ADDR
IPV4_VALUE = IPV4
IPV6 = IPV6ADDR
IPV6_VALUE = IPV6
IP_VALUE = "({IPV4ADDR}|{IPV6ADDR})".format(IPV4ADDR=IPV4ADDR, IPV6ADDR=IPV6ADDR)
IP_ADDR = IP_VALUE

LISTENING_PORT = "(\\d{1,5})"
IPV4_SERVICE = "({IPV4}:{LISTENING_PORT})".format(IPV4=IPV4, LISTENING_PORT=LISTENING_PORT)

IP = "(?P<ip>{IP_VALUE})".format(IP_VALUE=IP_VALUE)

# Datetime
YEAR_VALUE = "(\\d{4})"
MONTH_VALUE = "(\\d{2})"
DAY_VALUE = "(\\d{2})"
HOUR_VALUE = "(\\d{2})"
MINUTE_VALUE = "(\\d{2})"
SECOND_VALUE = "(\\d{2})"
MICROSECOND_VALUE = "(\\d{6})"
NANOSECOND_VALUE = "(\\d{9})"
MONTH_SHORT_NAME_VALUE = "(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"
DATE_SIMPLE_VALUE = "({YEAR_VALUE}{MONTH_VALUE}{DAY_VALUE})".format(YEAR_VALUE=YEAR_VALUE, MONTH_VALUE=MONTH_VALUE, DAY_VALUE=DAY_VALUE)
DATE_VALUE = "({YEAR_VALUE}-{MONTH_VALUE}-{DAY_VALUE})".format(YEAR_VALUE=YEAR_VALUE, MONTH_VALUE=MONTH_VALUE, DAY_VALUE=DAY_VALUE)
DATE_DMY_VALUE = "({DAY_VALUE}/{MONTH_SHORT_NAME_VALUE}/{YEAR_VALUE})".format(DAY_VALUE=DAY_VALUE, MONTH_SHORT_NAME_VALUE=MONTH_SHORT_NAME_VALUE, YEAR_VALUE=YEAR_VALUE)
DATE_YMD_VALUE = "{YEAR_VALUE}-{MONTH_VALUE}-{DAY_VALUE}".format(YEAR_VALUE=YEAR_VALUE, MONTH_VALUE=MONTH_VALUE, DAY_VALUE=DAY_VALUE)
TIME_SIMPLE_VALUE = "({HOUR_VALUE}{MINUTE_VALUE}{SECOND_VALUE})".format(HOUR_VALUE=HOUR_VALUE, MINUTE_VALUE=MINUTE_VALUE, SECOND_VALUE=SECOND_VALUE)
TIME_VALUE = "({HOUR_VALUE}:{MINUTE_VALUE}:{SECOND_VALUE})".format(HOUR_VALUE=HOUR_VALUE, MINUTE_VALUE=MINUTE_VALUE, SECOND_VALUE=SECOND_VALUE)
DATETIME_SIMPLE_VALUE = "({DATE_SIMPLE_VALUE}{TIME_SIMPLE_VALUE})".format(DATE_SIMPLE_VALUE=DATE_SIMPLE_VALUE, TIME_SIMPLE_VALUE=TIME_SIMPLE_VALUE)
DATETIME_VALUE = "({DATE_VALUE} {TIME_VALUE})".format(DATE_VALUE=DATE_VALUE, TIME_VALUE=TIME_VALUE)
TIMEZONE_VALUE = "([+|-]\\d{2}:?\\d{2})"

# Parameters
QUOTED_RAW_VALUE = "((\\\\\\\"|[^\\\"])*)"
QUOTED_VALUE = "(\"?{QUOTED_RAW_VALUE}\"?)".format(QUOTED_RAW_VALUE=QUOTED_RAW_VALUE)
NAME_VALUE = "([a-zA-Z_][a-zA-Z0-9-_]*)"
OPTION = "({NAME_VALUE}={QUOTED_VALUE})".format(NAME_VALUE=NAME_VALUE, QUOTED_VALUE=QUOTED_VALUE)

# Datetime with named groups
YEAR = "(?P<year>{YEAR_VALUE})".format(YEAR_VALUE=YEAR_VALUE)
MONTH = "(?P<month>{MONTH_VALUE})".format(MONTH_VALUE=MONTH_VALUE)
MONTH_SHORT_NAME = "(?P<month_short_name>{MONTH_SHORT_NAME_VALUE})".format(MONTH_SHORT_NAME_VALUE=MONTH_SHORT_NAME_VALUE)
DAY = "(?P<day>{DAY_VALUE})".format(DAY_VALUE=DAY_VALUE)
HOUR = "(?P<hour>{HOUR_VALUE})".format(HOUR_VALUE=HOUR_VALUE)
MINUTE = "(?P<minute>{MINUTE_VALUE})".format(MINUTE_VALUE=MINUTE_VALUE)
SECOND = "(?P<second>{SECOND_VALUE})".format(SECOND_VALUE=SECOND_VALUE)
MICROSECOND = "(?P<microsecond>{MICROSECOND_VALUE})".format(MICROSECOND_VALUE=MICROSECOND_VALUE)
NANOSECOND = "(?P<nanosecond>{NANOSECOND_VALUE})".format(NANOSECOND_VALUE=NANOSECOND_VALUE)
DATE_SIMPLE = "(?P<date_simple>{YEAR}{MONTH}{DAY})".format(YEAR=YEAR, MONTH=MONTH, DAY=DAY)
DATE = "(?P<date>{YEAR}-{MONTH}-{DAY})".format(YEAR=YEAR, MONTH=MONTH, DAY=DAY)
DATE_DMY = "(?P<date_dmy>{DAY}/{MONTH_SHORT_NAME}/{YEAR})".format(YEAR=YEAR, MONTH_SHORT_NAME=MONTH_SHORT_NAME, DAY=DAY)
DATE_YMD = "(?P<date_ymd>{YEAR}-{MONTH}-{DAY})".format(YEAR=YEAR, MONTH=MONTH, DAY=DAY)
TIME_SIMPLE = "(?P<time_simple>{HOUR}{MINUTE}{SECOND})".format(HOUR=HOUR, MINUTE=MINUTE, SECOND=SECOND)
TIME = "(?P<time>{HOUR}:{MINUTE}:{SECOND})".format(HOUR=HOUR, MINUTE=MINUTE, SECOND=SECOND)
DATETIME_SIMPLE = "(?P<datetime_simple>{DATE_SIMPLE}{TIME_SIMPLE})".format(DATE_SIMPLE=DATE_SIMPLE, TIME_SIMPLE=TIME_SIMPLE)
DATETIME = "(?P<datetime>{DATE} {TIME})".format(DATE=DATE, TIME=TIME)
TIMEZONE = "(?P<timezone>(?P<timezone_flag>[+|-])(?P<timezone_hour>\\d{2}):?(?P<timezone_minute>\\d{2}))"
ISO_FORMAT_TIME_VALUE = "({DAY_VALUE}/{MONTH_SHORT_NAME_VALUE}/{YEAR_VALUE}:{HOUR_VALUE}:{MINUTE_VALUE}:{SECOND_VALUE} {TIMEZONE})".format(
    DAY_VALUE=DAY_VALUE,
    MONTH_SHORT_NAME_VALUE=MONTH_SHORT_NAME_VALUE,
    YEAR_VALUE=YEAR_VALUE,
    HOUR_VALUE=HOUR_VALUE,
    MINUTE_VALUE=MINUTE_VALUE,
    SECOND_VALUE=SECOND_VALUE,
    TIMEZONE=TIMEZONE,
    )
ISO_FORMAT_TIME_WITH_MICROSECOND_VALUE = "({DAY_VALUE}/{MONTH_SHORT_NAME_VALUE}/{YEAR_VALUE}:{HOUR_VALUE}:{MINUTE_VALUE}:{SECOND_VALUE}(\\.{MICROSECOND_VALUE})? {TIMEZONE})".format(
    DAY_VALUE=DAY_VALUE,
    MONTH_SHORT_NAME_VALUE=MONTH_SHORT_NAME_VALUE,
    YEAR_VALUE=YEAR_VALUE,
    HOUR_VALUE=HOUR_VALUE,
    MINUTE_VALUE=MINUTE_VALUE,
    SECOND_VALUE=SECOND_VALUE,
    MICROSECOND_VALUE=MICROSECOND_VALUE,
    TIMEZONE=TIMEZONE,
    )
ISO_FORMAT_TIME = "((?P<datetime>{DATE_DMY}:{TIME} {TIMEZONE}))".format(
    DATE_DMY=DATE_DMY,
    TIME=TIME,
    TIMEZONE=TIMEZONE,
    )
ISO_FORMAT_TIME_WITH_MICROSECOND = "((?P<datetime>{DATE_DMY}:{TIME}\\.{MICROSECOND} {TIMEZONE}))".format(
    DATE_DMY=DATE_DMY,
    TIME=TIME,
    MICROSECOND=MICROSECOND,
    TIMEZONE=TIMEZONE,
    )
ISO_FORMAT_TIME_WITH_NANOSECOND = "((?P<datetime>{DATE_DMY}:{TIME}\\.{NANOSECOND} {TIMEZONE}))".format(
    DATE_DMY=DATE_DMY,
    TIME=TIME,
    NANOSECOND=NANOSECOND,
    TIMEZONE=TIMEZONE,
    )
LOG_LEVEL = "(?P<log_level>debug|info|warn|warning|err|error|critical|notice|DEBUG|INFO|NOTICE|WARN|WARNING|ERR|ERROR|CRITICAL)"


# HTTP
HTTP_REQUEST = "(?P<request>(?P<request_method>.+) (?P<request_path>[^\\?]+)(\\?(?P<request_query>.+))? (?P<http_version>.+))"
DEFAULT_NGINX_ACCESS = "(?P<remote_addr>.+) (?P<server_name>.+) (?P<remote_user>.+) \\[(?P<time_local>.+)\\] \\\"{HTTP_REQUEST}\\\" (?P<status>\\d+) (?P<body_bytes_sent>\\d+) \\\"(?P<http_referer>.+)\\\" \\\"(?P<http_user_agent>.+)\\\" \\\"(?P<http_x_forwarded_for>.+)\\\""
SIMPLE_NGINX_ACCESS = "{DEFAULT_NGINX_ACCESS} (?P<upstream_status>.+) (?P<upstream_addr>.+) (?P<request_time>.+) (?P<upstream_response_time>.+) \\\"(?P<request_body>.+)\\\""

# OTHER
MESSAGE = "(?P<message>.*)"


import re
import copy
import string


def get_default_rules():
    """Get default rule sets.
    """
    rules = {}
    _globals = dict(globals())
    for key, value in _globals.items():
        if not key.startswith("_") and key[0] in string.ascii_uppercase:
            rules[key] = value
    return expend_vars(rules)


def expend_vars(vars):
    """Expand with variables.

    Returns
        return expanded vars.

    Raise exception
        If some variable expand failed.
    """
    vars = copy.copy(vars)
    VAR = "(?P<var>\\{[A-Z_][A-Z_0-9]*\\})"
    while True:
        all_expended_flag = True
        for key in vars.keys():
            value = vars[key]
            names = re.findall(VAR, value)
            if names:
                all_expended_flag = False
                vars[key] = value.format(**vars)
        if all_expended_flag:
            break
    return vars




# -*- coding: utf-8 -*-
import os
from io import open
import click
import yaml
import logging

from fastutils import logutils

from .login import LogInput
from .logfilter import LogFilter
from .logout import LogOutput
from .logparser import LogParser

DEFAULT_LOGGING_SETTINGS = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "%(asctime)s\t%(process)d\t%(thread)d\t%(levelname)s\t%(pathname)s\t%(lineno)d\t%(message)s",
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
        },
        "logfile": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "filename": "server.log",
            "maxBytes": 1024*1024*128,
            "backupCount": 36,
            "encoding": "utf-8",
        }
    },
    "loggers": {
    },
    "root": {
        "level": "INFO",
        "handlers": ["console", "logfile"],
        "propagate": True,
    }
}

logger = logging.getLogger(__name__)

def deep_merge(data1, data2):
    for key2, value2 in data2.items():
        value1 = data1.get(key2, None)
        if isinstance(value1, dict) and isinstance(value2, dict):
            deep_merge(value1, value2)
        else:
            data1[key2] = value2
        

def load_input(settings):
    input_settings = settings.get("input", {}) or {}
    try:
        input_object = LogInput.init(input_settings)
        if not input_object:
            logger.error(u"Error: input data type error.")
            os.sys.exit(1)
        return input_object
    except RuntimeError as err:
        logger.error(u"Error: init input failed, message={message}.".format(message=str(err)))
        os.sys.exit(2)

def load_output(settings):
    output_setting = settings.get("output", {}) or {}
    try:
        output = LogOutput.init(output_setting)
        return output
    except Exception as err:
        logger.error("Error: init output failed, message={message}.".format(message=str(err)))
        os.sys.exit(2)

def load_parser(settings):
    parser_setting = settings.get("parser", {}) or {}
    if not parser_setting:
        return None
    try:
        parser = LogParser.init(parser_setting)
        return parser
    except Exception as err:
        logger.error("Error: init parser failed, message={message}.".format(message=str(err)))
        os.sys.exit(1)

def load_filters(settings):
    filters = []
    filters_setting = settings.get("filters", []) or []
    if filters_setting:
        for filter_setting in filters_setting:
            try:
                filter = LogFilter.init(filter_setting)
                if not filter:
                    logger.error("Error: filter config error.")
                filters.append(filter)
            except RuntimeError as error:
                logger.error("Error: init filter failed, message={message}。".format(message=str(error)))
                os.sys.exit(1)
    return filters

@click.group()
@click.option("-c", "--config", required=True, help=u"Config file path. The config file must in yaml format.")
@click.pass_context
def main(ctx, config):
    """Parse log file as input and export the data to database as output.
    """
    # read settings from config file
    with open(config, "r", encoding="utf-8") as fobj:
        settings = yaml.safe_load(fobj)
    if not settings:
        settings = {}
    
    # setup logging
    logutils.setup(**settings)

    # other settings
    ctx.ensure_object(dict)
    ctx.obj["config"] = config
    ctx.obj["settings"] = settings

@main.command()
@click.pass_context
def server(ctx):
    """Start log handler server.
    """
    settings = ctx.obj["settings"]
    input = load_input(settings)
    output_object = load_output(settings)
    parser = load_parser(settings)
    filters = load_filters(settings)
    input.set_handler(output_object, filters=filters, parser=parser)
    input.loop()

@main.command()
@click.argument("text", nargs=1, required=True)
@click.pass_context
def test(ctx, text):
    """Parse the example text and print out parse result.
    """
    settings = ctx.obj["settings"]
    parser = load_parser(settings)
    if parser:
        info = parser.line_handler(text, keep_not_matched_lines=False)
        text = yaml.dump(info, allow_unicode=True)
        print(text)
    else:
        print("No parser configed...")

@main.command()
@click.option("-i", "--include-blank-lines", is_flag=True, help="Parse empty lines or not. Default to NO parse.")
@click.option("-e", "--encoding", default="utf-8", help="Encoding of the input. Default to UTF-8.")
@click.argument("filename", nargs=1, required=False)
@click.pass_context
def scan(ctx, include_blank_lines, encoding, filename):
    """Try example settings on a test file.
    """
    settings = ctx.obj["settings"]
    scan_input_settings = {
        "type": "file",
        "ignore-blank-lines": not include_blank_lines,
        "filename": filename,
        "encoding": encoding,
    }
    settings["input"] = scan_input_settings
    input_object = load_input(settings)
    output = load_output(settings)
    parser = load_parser(settings)
    filters = load_filters(settings)
    input_object.set_handler(output, filters=filters, parser=parser)
    input_object.loop()

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from __future__ import print_function
from io import open
import re
import copy
import logging
from pprint import pprint

import requests
from fastutils import dictutils

logger = logging.getLogger(__name__)

LOG_OUTPUT_TYPES = {}

def register_output_type(name, klass):
    LOG_OUTPUT_TYPES[name] = klass

def unregister_output_type(name):
    del LOG_OUTPUT_TYPES[name]

def get_output_class(name):
    return LOG_OUTPUT_TYPES.get(name, None)

class LogOutput(object):

    def __init__(self, settings):
        self.settings = settings or {}

    def update(self, info):
        """True means saved, False means cached, raise exception means something error.
        """
        raise NotImplementedError()

    def flush(self):
        """Force to save the data.
        """
        raise NotImplementedError()

    @classmethod
    def init(cls, settings):
        output_type = settings.get("type", "mysql")
        output_class = get_output_class(output_type)
        if not output_class:
            return None
        return output_class(settings)

class BufferedLogOutput(LogOutput):

    def __init__(self, settings):
        super().__init__(settings)
        self.buffer_size = self.settings.get("buffer-size", 10) or 10
        self.count = 0
        self.buffer = []

    def update(self, info):
        self.buffer.append(info)
        self.count += 1
        if self.count % self.buffer_size == 0:
            self.buffer_handler()
            return True
        else:
            return False

    def flush(self):
        self.buffer_handler()
        return True

    def buffer_handler(self):
        for info in self.buffer:
            self.line_handler(info)
        self.buffer = []
        return True

    def line_handler(self, info):
        raise NotImplementedError()

class LogToStdout(BufferedLogOutput):

    def __init__(self, settings):
        super().__init__(settings)
        self.output_template = self.settings.get("output-template", None)

    def line_handler(self, info):
        if self.output_template:
            print(self.output_template.format(**info))
        else:
            pprint(info)

class LogToMysql(BufferedLogOutput):

    DEFAULT_MYSQL_SETTINGS = {
        "autocommit": True,
        "charset": "utf8",
    }

    def __init__(self, settings):
        super(LogToMysql, self).__init__(settings)
        # mysql properties
        self.mysql_settings = copy.copy(self.DEFAULT_MYSQL_SETTINGS)
        self.mysql_settings.update(self.settings.get("mysql", {}) or {})
        self.mysql_connect = None
        # properties
        self.ignore_not_matched_lines = self.settings.get("ignore-not-matched-lines", True)
        self.keep_failed_lines = self.settings.get("keep-failed-lines", "") or ""
        self.inserts = self.settings.get("inserts", {}) # settings.inserts required

    def do_keep_failed_lines(self, lines):
        if self.keep_failed_lines:
            with open(self.keep_failed_lines, "a", encoding="utf-8") as fobj:
                for line in lines:
                    fobj.write(u"{0}\n".format(line))

    def get_cursor(self):
        import MySQLdb as mysql
        if not self.mysql_connect:
            self.mysql_connect = mysql.connect(**self.mysql_settings)
        else:
            self.mysql_connect.ping()
        return self.mysql_connect.cursor()

    def get_categoryed_buffer(self):
        data = {"DEFAULT": []}
        for info in self.buffer:
            if not info["_matched"] and self.ignore_not_matched_lines:
                continue
            matched_name = info["_matched_name"]
            if matched_name in self.inserts:
                if matched_name in data:
                    data[matched_name].append(info)
                else:
                    data[matched_name] = [info]
            else:
                data["DEFAULT"].append(info)
        return data

    def buffer_handler(self):
        if self.buffer:
            try:
                cursor = self.get_cursor()
            except Exception as err:
                logger.exception("Error: Connect to mysql failed, message={0}。".format(str(err)))
                cursor = None
                self.do_keep_failed_lines(self.buffer)
            if cursor:
                for insert_name, rows in self.get_categoryed_buffer().items():
                    if insert_name in self.inserts:
                        sql_settings = self.inserts[insert_name]
                        sql_template = sql_settings["sql"]
                        sql_fields = sql_settings["fields"]
                        rows = self.get_insert_rows(rows, sql_fields)
                        try:
                            cursor.executemany(sql_template, rows)
                            logger.info("Info: insert data into mysql success, {0} records inserted.".format(len(rows)))
                        except Exception as err:
                            logger.exception("Error: insert data into mysql failed, message={0}, sql_template={1}。".format(str(err), sql_template))
                            self.do_keep_failed_lines(self.buffer)
        self.buffer = []
        return True

    def get_insert_rows(self, rows, fields):
        data = []
        for row in rows:
            row_data = []
            for field_name in fields:
                row_data.append(row[field_name])
            data.append(row_data)
        return data

    def fix_missing_field(self, sql_template, rows):
        fields = re.findall("\\%\\(([a-zA-Z0-9_]+)\\)s", sql_template)
        for row in rows:
            for field in fields:
                if not field in row:
                    row[field] = None

class PrintNotMatchedLine(BufferedLogOutput):

    def line_handler(self, info):
        matched = info.get("_matched", False)
        if not matched:
            print(info["_line"])

class LogToHttp(LogOutput):

    def __init__(self, settings):
        super().__init__(settings)
        self.server = settings.get("server")
        self.method = settings.get("method", "GET")
        self.headers = settings.get("headers", {})
        self.extra_params = settings.get("extra-params", {})
        self.params_selector = settings.get("params", None)
        self.extra_data = settings.get("extra-data", {})
        self.data_selector = settings.get("data", None)
        self.extra_json = settings.get("extra-json", {})
        self.json_selector = settings.get("json", None)
    
    def update(self, info):
        """True means saved, False means cached, raise exception means something error.
        """
        print(info)
        params = {}
        params.update(self.extra_params)
        if self.params_selector:
            more_params = dictutils.select(info, self.params_selector, {})
            if more_params:
                params.update(more_params)
        if not params:
            params = None
        
        data = {}
        data.update(self.extra_data)
        if self.data_selector:
            more_data = dictutils.select(info, self.data_selector, {})
            if more_data:
                data.update(more_data)
        if not data:
            data = None
        
        json = {}
        json.update(self.extra_json)
        if self.json_selector:
            more_json = dictutils.select(info, self.json_selector, {})
            if more_json:
                json.update(more_json)
        if not json:
            json = None

        response = requests.request(self.method, self.server, params=params, data=data, json=json, verify=False)
        logger.info("{} {}".format(response.status_code, response.text))

    def flush(self):
        """Force to save the data.
        """
        pass

register_output_type("mysql", LogToMysql)
register_output_type("stdout", LogToStdout)
register_output_type("print-not-matched-line", PrintNotMatchedLine)
register_output_type("http", LogToHttp)



# -*- coding: utf-8 -*-
from io import open
import re
import json
import copy
import urllib
import datetime
import logging
from .rules import get_default_rules
from .rules import expend_vars

logger = logging.getLogger(__name__)

LOG_PARSER_TYPES = {}

def register_parser_type(name, klass):
    LOG_PARSER_TYPES[name] = klass

def unregister_parser_type(name):
    del LOG_PARSER_TYPES[name]

def get_parser_class(name):
    return LOG_PARSER_TYPES.get(name, None)

LOG_TRANSFORMERS = {}

def register_transformer_type(name, klass):
    LOG_TRANSFORMERS[name] = klass

def unregister_transformer_type(name):
    del LOG_TRANSFORMERS[name]

def get_transformer_class(name):
    return LOG_TRANSFORMERS.get(name, None)

class LogParser(object):

    def __init__(self, settings):
        self.settings = settings or {}
        # properties
        self.keep_not_matched_lines = self.settings.get("keeep-not-matched-lines", "") or ""
        self.transforms = self.settings.get("transforms", []) or []

    def do_keep_not_matched_line(self, info):
        with open(self.keep_not_matched_lines, "a", encoding="utf-8") as fobj:
            fobj.write(u"{0}\n".format(info["_line"]))
        logger.warn(u"Parse line failed: {0}".format(info["_line"]))

    @classmethod
    def init(cls, settings):
        parser_type = settings.get("type", "regex")
        parser_class = get_parser_class(parser_type)
        if not parser_class:
            return None
        return parser_class(settings)

    def line_handler(self, line):
        info = self.parse_line(line)
        if self.transforms:
            self.do_transform(info)
        return info

    def parse_line(self, line):
        return {
            "_line": line,
        }

    def do_transform(self, data):
        for transform_settings in self.transforms:
            try:
                field = transform_settings.get("field", None)
                transform_type = transform_settings.get("type", None)
                if transform_type:
                    transformer_class = get_transformer_class(transform_type)
                    if transformer_class:
                        data[field] = transformer_class(transform_settings).do_transform(data, field)
            except Exception:
                logger.exception("do transform failed.")

class Transformer(object):
    """Base class of Transformer.
    """

    def __init__(self, settings):
        self.settings = settings or {}

    def do_transform(self, data, key):
        return data[key]

class StrftimeTransformer(Transformer):
    """Load datetime string into datetime object or timestamp value.
    """

    def __init__(self, settings):
        super().__init__(settings)
        self.strptime_template = self.settings.get("strptime", "%Y-%m-%d %H:%M:%S")
        self.strftime_template = self.settings.get("strftime", "%Y-%m-%d %H:%M:%S")
        self.use_timestamp = self.settings.get("timestamp", False)
    
    def do_transform(self, data, field):
        value = data[field]
        value = datetime.datetime.strptime(value, self.strptime_template)
        if self.use_timestamp:
            return value.timestamp()
        else:
            return value.strftime(self.strftime_template)

class Str2Number(Transformer):
    """Turn str digits to int value.
    """

    def do_transform(self, data, field):
        value = data[field]
        if "." in value:
            return float(value)
        else:
            return int(value)

class ForceInt(Transformer):
    """Turn str digits to int value, and turn float value to int value.
    """

    def do_transform(self, data, field):
        value = data[field]
        return int(value)

class ForceFloat(Transformer):
    def do_transform(self, data, field):
        value = data[field]
        return float(value)

class FormulaCompute(Transformer):

    def __init__(self, settings):
        super().__init__(settings)
        self.formula = self.settings.get("formula", "0")

    def do_transform(self, data, key):
        from sympy import sympify
        formula = self.formula.format(**data)
        return sympify(formula)

class RequestBody(Transformer):
    """Load nginx request_body data.

    -
        Returns None

    a=b&c=d
        Returns {
            "a": "b",
            "c": "d"
        }
    
    "{\x0A\x09\x22a\x22:\x22b\x22,\x0A\x09\x22c\x22:\x22d\x22}"
        Returns {
            "a": "b",
            "c": "d"
        }

    """
    def __init__(self, settings):
        super().__init__(settings)
    
    def do_transform(self, data, key):
        value = data[key]
        if value == "-":
            return None
        else:
            try:
                return json.loads(eval(f"""b"{value}" """))
            except Exception:
                try:
                    return dict(urllib.parse.parse_qsl(value))
                except Exception:
                    return False

class JsonParser(LogParser):

    def parse_line(self, line, keep_not_matched_lines=None):
        try:
            data = json.loads(line)
            if isinstance(data, dict):
                data["_matched"] = True
            else:
                data = {
                    "_data": data,
                    "_matched": False,
                }
        except json.JSONDecodeError:
            data = {}
            data["_matched"] = False
        data["_line"] = line
        return data

class RegexParser(LogParser):
    
    def __init__(self, settings):
        super(RegexParser, self).__init__(settings)
        # properties
        self.use_default_rules = self.settings.get("use-default-rules", True)
        self.rules = self.get_rules()
        self.matches = self.get_matches()

    def get_rules(self):
        rule_map = {}
        if self.use_default_rules:
            rule_map.update(get_default_rules())
        rule_map.update(self.settings.get("rules", {}) or {})
        return expend_vars(rule_map)

    def get_matches(self):
        raw_matches = self.settings.get("matches", {})
        rule_map = copy.copy(self.rules)
        rule_map.update(raw_matches)
        rule_map = expend_vars(rule_map)
        match_items = []
        for match_name in raw_matches.keys():
            match_pattern = rule_map[match_name]
            match_items.append((match_name, match_pattern))
        return match_items

    def parse_line(self, line, keep_not_matched_lines=None):
        if keep_not_matched_lines is None:
            keep_not_matched_lines = self.keep_not_matched_lines
        data = {
            "_line": line,
        }
        matched = False
        if self.matches:
            for match_name, match_pattern in self.matches:
                try:
                    match = re.match(match_pattern, line)
                    if match:
                        matched = True
                        data["_matched_name"] = match_name
                        data.update(match.groupdict())
                        break
                except Exception as err:
                    msg = u"Error: regex match failed, match_name={0}，match_pattern={1}，error={2}。".format(
                        match_name,
                        match_pattern,
                        str(err),
                    )
                    logger.error(msg)
                    raise err
        data["_matched"] = matched
        if not matched and keep_not_matched_lines:
            self.do_keep_not_matched_line(data)
        return data


register_transformer_type("strftime", StrftimeTransformer)
register_transformer_type("str2number", Str2Number),
register_transformer_type("formula", FormulaCompute)
register_transformer_type("int", ForceInt)
register_transformer_type("float", ForceFloat)
register_transformer_type("request_body", RequestBody)

register_parser_type("regex", RegexParser)
register_parser_type("json", JsonParser)



# -*- coding: utf-8 -*-
import re
import unittest
from logio.rules import *
from logio import rules
from logio.logparser import expend_vars
from logio.logparser import get_default_rules

class TestLogParse(unittest.TestCase):

    def test01(self):
        """test year
        """
        for x in range(1900, 2999):
            year = str(x)
            assert re.match(YEAR, year).groupdict()["year"] == year

    def test02(self):
        """test month
        """
        for x in range(1, 13):
            month = "{:02d}".format(x)
            assert re.match(MONTH, month).groupdict()["month"] == month

    def test03(self):
        """test day
        """
        for x in range(1, 32):
            day = "{:02d}".format(x)
            assert re.match(DAY, day).groupdict()["day"] == day

    def test04(self):
        """test hour
        """
        for x in range(1, 25):
            hour = "{:02d}".format(x)
            assert re.match(HOUR, hour).groupdict()["hour"] == hour

    def test05(self):
        """test minute
        """
        for x in range(1, 61):
            minute = "{:02d}".format(x)
            assert re.match(MINUTE, minute).groupdict()["minute"] == minute

    def test06(self):
        """test second
        """
        for x in range(1, 61):
            second = "{:02d}".format(x)
            assert re.match(SECOND, second).groupdict()["second"] == second

    def test07(self):
        """test date_simple, e.g. 20190630
        """
        info = re.match(DATE_SIMPLE, "20190630").groupdict()
        assert info["year"] == "2019"
        assert info["month"] == "06"
        assert info["day"] == "30"
        assert info["date_simple"] == "20190630"

    def test08(self):
        """test DATE, e.g. 2019-06-30
        """
        info = re.match(DATE, "2019-06-30").groupdict()
        assert info["year"] == "2019"
        assert info["month"] == "06"
        assert info["day"] == "30"

    def test09(self):
        """test DATE_DMY, e.g. 03/Jun/2019
        """
        for month_short_name in "Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec".split("|"):
            info = re.match(DATE_DMY, "01/{month_short_name}/2019".format(month_short_name=month_short_name)).groupdict()
            assert info["year"] == "2019"
            assert info["month_short_name"] == month_short_name
            assert info["day"] == "01"

    def test10(self):
        """test TIME_SIMPLE, e.g. 200103
        """
        info = re.match(TIME_SIMPLE, "200103").groupdict()
        assert info["hour"] == "20"
        assert info["minute"] == "01"
        assert info["second"] == "03"
        assert info["time_simple"] == "200103"
    
    def test11(self):
        """test TIME, e.g. 20:01:03
        """
        info = re.match(TIME, "20:01:03").groupdict()
        assert info["hour"] == "20"
        assert info["minute"] == "01"
        assert info["second"] == "03"
        assert info["time"] == "20:01:03"
        
    def test12(self):
        """test DATETIME_SIMPLE, e.g. 20190630200501
        """
        info = re.match(DATETIME_SIMPLE, "20190630200501").groupdict()
        assert info["year"] == "2019"
        assert info["month"] == "06"
        assert info["day"] == "30"
        assert info["date_simple"] == "20190630"
        assert info["hour"] == "20"
        assert info["minute"] == "05"
        assert info["second"] == "01"
        assert info["time_simple"] == "200501"
        assert info["datetime_simple"] == "20190630200501"

    def test13(self):
        """test DATETIME, e.g. 2019-06-30 20:05:01
        """
        info = re.match(DATETIME, "2019-06-30 20:05:01").groupdict()
        assert info["year"] == "2019"
        assert info["month"] == "06"
        assert info["day"] == "30"
        assert info["date"] == "2019-06-30"
        assert info["hour"] == "20"
        assert info["minute"] == "05"
        assert info["second"] == "01"
        assert info["time"] == "20:05:01"
        assert info["datetime"] == "2019-06-30 20:05:01"
     
    def test14(self):
        """test TIMEZONE, e.g. +0800
        """
        info = re.match(TIMEZONE, "+0800").groupdict()
        assert info["timezone"] == "+0800"

    def test15(self):
        """test TIMEZONE_ISO, e.g. +08:06
        """
        info = re.match(TIMEZONE, "+08:06").groupdict()
        assert info["timezone"] == "+08:06"

    def test16(self):
        """test QUOTED_VALUE, e.g. 
        """
        for value in [
                "abc",
                "hello world",
                "http://a.com",
                '''"es-ES,es;q=0.8"''',
                '''"Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.11 (KHTML, like Gecko) Chrome/23.0.1271.97 Safari/537.11"''',
                "it\\\"s not ok",
                ]:
            info = re.match(QUOTED_VALUE, value).group()
            assert info == value

    def test17(self):
        """test NAME_VALUE, e.g. filename
        """
        for name in ["filename", "op"]:
            info = re.match(NAME_VALUE, name).group()
            assert info == name
    
    def test19(self):
        """test OPTION
        """
        for option in [
                "a=123",
                "a=hello world",
                '''a="hello world"''',
                '''a="it\\\"s not ok"''',
                ]:
            info = re.match(OPTION, option).group()
            assert info == option

    def test20(self):
        """test IPV4
        """
        for ip in ["127.0.0.1", "10.0.0.1", "114.114.114.114"]:
            info = re.match(IPV4, ip).group()
            assert info == ip

    def test21(self):
        """test access time, e.g. [03/Jan/2013:21:17:20 -0600]
        """
        for time in ["[03/Jan/2013:21:17:20 -0600]"]:
            info = re.match(ISO_FORMAT_TIME_VALUE, time).group()
            assert info == time

    def test22(self):
        """test access time, e.g. [03/Jan/2013:21:17:20 -0600]
        """
        info = re.match(ISO_FORMAT_TIME, "[03/Jan/2013:21:17:20 -0600]").groupdict()
        assert info["year"] == "2013"
        assert info["month_short_name"] == "Jan"
        assert info["day"] == "03"
        assert info["hour"] == "21"
        assert info["minute"] == "17"
        assert info["second"] == "20"
        assert info["timezone"] == "-0600"
        assert info["datetime"] == "03/Jan/2013:21:17:20 -0600"

    def test23(self):
        """test IPV4_SERVICE
        """
        info = re.match(IPV4_SERVICE, "127.0.0.1:9000").group()
        assert info == "127.0.0.1:9000"
    
    def test30(self):
        """test expend_vars
        """
        rules = {
            "YEAR": YEAR,
            "MONTH": MONTH,
            "DAY": DAY,
            "HOUR": HOUR,
            "MINUTE": MINUTE,
            "SECOND": SECOND,
            "DATE_SIMPLE": "(?P<date_simple>{YEAR}{MONTH}{DAY})",
            "TIME_SIMPLE": "(?P<time_simple>{HOUR}{MINUTE}{SECOND})",
            "DATETIME_SIMPLE": "(?P<datetime>{DATE_SIMPLE}{TIME_SIMPLE})",
        }
        new_rules = expend_vars(rules)
        assert not "YEAR" in new_rules["DATE_SIMPLE"]
        assert not "MONTH" in new_rules["DATE_SIMPLE"]
        assert not "DAY" in new_rules["DATE_SIMPLE"]
        assert not "HOUR" in new_rules["TIME_SIMPLE"]
        assert not "MINUTE" in new_rules["TIME_SIMPLE"]
        assert not "SECOND" in new_rules["TIME_SIMPLE"]
        assert not "DATE_SIMPLE" in new_rules["DATETIME_SIMPLE"]
        assert not "TIME_SIMPLE" in new_rules["DATETIME_SIMPLE"]

    def test31(self):
        rules = get_default_rules()
        assert rules

    def test32(self):
        text = "[03/Jul/2019:14:41:43.570635527 +0800] conn=16959994 op=1 RESULT err=0 tag=101 nentries=1 etime=0.0001746563"
        rules = get_default_rules()
        rules.update({"RESULT": "({ISO_FORMAT_TIME_WITH_MICROSECOND} conn=(?P<conn>\\d+) op=(?P<op>\\d+).*)"})
        rules = expend_vars(rules)
        print(rules["RESULT"])
        match = re.match(rules["RESULT"], text)
        print(match)
        assert match

    def test33(self):
        for name in dir(rules):
            if not name.startswith("_"):
                pattern = getattr(rules, name)
                assert re.compile(pattern)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import os
import io
from io import open
import time
import codecs
import datetime
import threading
import logging
from tail import TailReader

logger = logging.getLogger(__name__)

def get_stdin_reader(encoding=None):
    if hasattr(os.sys.stdin, "reconfigure"):
        os.sys.stdin.reconfigure(encoding=encoding)
        return os.sys.stdin
    else:
        os.sys.stdin = codecs.getreader(encoding)(os.sys.stdin)
        return os.sys.stdin


LOG_INPUT_TYPES = {}

def register_input_type(name, klass):
    LOG_INPUT_TYPES[name] = klass

def unregister_input_type(name):
    del LOG_INPUT_TYPES[name]

def get_input_class(name):
    return LOG_INPUT_TYPES.get(name, None)

class LogInput(object):

    def __init__(self, settings):
        self.settings = settings or {}
        self.parser = None
        self.output = None
        self.filters = None
        # properties
        self.ignore_blank_lines = self.settings.get("ignore-blank-lines", True)
        self.encoding = self.settings.get("encoding", "utf-8")
        self.report_interval = self.settings.get("report-interval", 1)
        # private properties
        self.lines_handlered = 0
        self.last_reported_timestamp = time.time()
        # self.report_thread = threading.Thread(target=self.report)
        # self.report_thread.setDaemon(True)
        # self.report_thread.start()

    def set_handler(self, output, filters=None, parser=None):
        self.parser = parser
        self.output = output
        self.filters = filters

    def loop(self):
        raise NotImplementedError()

    def report(self):
        while True:
            time.sleep(self.report_interval)
            msg = "{0} lines handlered.".format(self.lines_handlered)
            logger.info(msg)

    def line_handler(self, line):
        self.lines_handlered += 1
        if self.filters:
            for filter in self.filters:
                if filter.line_filter(line) == False:
                    return
        if self.parser:
            line_info = self.parser.line_handler(line)
        else:
            line_info = {
                "_line": line,
            }
        return self.output.update(line_info)
    
    def flush_output(self):
        return self.output.flush()

    @classmethod
    def init(cls, settings):
        input_type = settings.get("type", "stdin")
        input_class = get_input_class(input_type)
        if not input_class:
            return None
        return input_class(settings)

class StdinInput(LogInput):

    def loop(self):
        reader = get_stdin_reader(self.encoding)
        for line in reader.readlines():
            line = line.rstrip(u"\r\n")
            if line or not self.ignore_blank_lines:
                self.line_handler(line)

class FileInput(LogInput):

    def __init__(self, settings):
        super().__init__(settings)
        # properties
        self.filename = self.settings["filename"]

    def loop(self):
        if self.filename:
            reader = io.open(self.filename, encoding=self.encoding)
        else:
            reader = get_stdin_reader(encoding=self.encoding)
        for line in reader.readlines():
            line = line.rstrip(u"\r\n")
            if line or not self.ignore_blank_lines:
                self.line_handler(line)

class TailInput(LogInput):

    def __init__(self, settings):
        super().__init__(settings)
        # properties
        self.filename = self.settings["filename"]
        self.offset_file = self.settings.get("offset-file", None)
        self.read_from_end = self.settings.get("read-from-end", False)
        self.backup_patterns = self.settings.get("backup-patterns", None)
        self.sleep_interval = self.settings.get("sleep-interval", 1)
        self.blocking = self._get_blockeding_setting()

    def _get_blockeding_setting(self):
        result = None
        if "non-blocking" in self.settings:
            result = not self.settings["non-blocking"]
        if "blocking" in self.settings:
            result = self.settings["blocking"]
        if result is None:
            result = True
        return result

    def loop(self):
        filereader = TailReader(
            self.filename,
            self.offset_file,
            self.read_from_end,
            self.encoding,
            self.backup_patterns,
            )
        while True:
            c = 0
            for line in filereader.readlines():
                line = line.rstrip(u"\r\n")
                if line or not self.ignore_blank_lines:
                    stored = self.line_handler(line)
                    c += 1
                    if stored:
                        filereader.update_offset()
            stored = self.flush_output()
            if stored:
                filereader.update_offset()
            if not self.blocking:
                break
            if c < 1:
                time.sleep(self.sleep_interval)


register_input_type("stdin", StdinInput)
register_input_type("file", FileInput)
register_input_type("tail", TailInput)
